@@ -1,11 +1,49 @@
 #include "checksum.h"
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 
 #include "partclone.h" // for log_mesg() & cmd_opt
+#include "xxhash.h"
+
+#ifdef HAVE_ISAL
+#include <isa-l.h>
+#endif
 
 #define CRC32_SEED 0xFFFFFFFF
 
 static uint32_t crc_tab32[256] = { 0 };
 static int cs_mode = CSM_NONE;
+static XXH64_state_t* xxh64_state = NULL;
+
+/**
+ * ISA-L compatible CRC32 wrapper
+ * This function ensures bit-for-bit compatibility with the existing CRC32
+ * by properly handling the seed inversion as per IEEE 802.3 specification
+ */
+static inline uint32_t crc32_isa_compatible(uint32_t seed, void* buffer, long size) {
+#ifdef HAVE_ISAL
+    // ISA-L crc32_gzip_refl expects the seed to be bit-inverted
+    uint32_t ext_seed = ~seed;
+    uint32_t ext_crc = crc32_gzip_refl(ext_seed, (const unsigned char*)buffer, (uint64_t)size);
+    // Return bit-inverted result to match expected output
+    return ~ext_crc;
+#else
+    // Fallback to original implementation
+    unsigned char * buf = (unsigned char *)buffer;
+    const unsigned char * end = buf + size;
+    uint32_t tmp, long_c, crc = seed;
+
+    while (buf != end) {
+        /// update crc
+        long_c = *(buf++);
+        tmp = crc ^ long_c;
+        crc = (crc >> 8) ^ crc_tab32[tmp & 0xff];
+    };
+
+    return crc;
+#endif
+}
 
 unsigned get_checksum_size(int checksum_mode, int debug) {
 
@@ -18,6 +56,9 @@ unsigned get_checksum_size(int checksum_mode, int debug) {
 	case CSM_CRC32_0001:
 		return 4;
 
+	case CSM_XXH64:
+		return sizeof(XXH64_hash_t);
+
 	default:
 		log_mesg(0, 1, 1, debug, "Unknown checksum mode [%d]\n", checksum_mode);
 		return UINT_LEAST32_MAX;
@@ -35,6 +76,9 @@ const char *get_checksum_str(int checksum_mode) {
 	case CSM_CRC32:
 		return "CRC32";
 
+	case CSM_XXH64:
+		return "XXH64";
+
 	case CSM_CRC32_0001:
 		return "CRC32_0001";
 
@@ -80,6 +124,7 @@ void init_crc32(uint32_t* seed) {
  */
 void init_checksum(int checksum_mode, unsigned char* seed, int debug) {
 
+	cs_mode = checksum_mode;
 	switch(checksum_mode) {
 
 	case CSM_CRC32:
@@ -90,6 +135,13 @@ void init_checksum(int checksum_mode, unsigned char* seed, int debug) {
 		init_crc32((uint32_t*)seed);
 		break;
 
+	case CSM_XXH64:
+		if (xxh64_state == NULL) {
+			xxh64_state = XXH64_createState();
+		}
+		XXH64_reset(xxh64_state, 0); // Using 0 as seed
+		break;
+
 	case CSM_NONE:
 		// Nothing to do
 		// Leave seed alone as it may be NULL or point to a zero-sized array
@@ -99,8 +151,6 @@ void init_checksum(int checksum_mode, unsigned char* seed, int debug) {
 		log_mesg(0, 1, 1, debug, "Unknown checksum mode [%d]\n", checksum_mode);
 		break;
 	}
-
-	cs_mode = checksum_mode;
 }
 
 /// the crc32 function, reference from libcrc.
@@ -109,19 +159,7 @@ void init_checksum(int checksum_mode, unsigned char* seed, int debug) {
 /// http://www.lammertbies.nl/comm/info/nl_crc-calculation.html
 /// generate crc32 code
 uint32_t crc32(uint32_t seed, void* buffer, long size) {
-
-	unsigned char * buf = (unsigned char *)buffer;
-	const unsigned char * end = buf + size;
-	uint32_t tmp, long_c, crc = seed;
-
-	while (buf != end) {
-		/// update crc
-		long_c = *(buf++);
-		tmp = crc ^ long_c;
-		crc = (crc >> 8) ^ crc_tab32[tmp & 0xff];
-	};
-
-	return crc;
+    return crc32_isa_compatible(seed, buffer, size);
 }
 
 /**
@@ -156,18 +194,18 @@ static uint32_t crc32_0001(uint32_t seed, void* buffer, int size) {
  */
 void update_checksum(unsigned char* checksum, char* buf, int size) {
 
-	uint32_t* crc;
-
 	switch(cs_mode)
 	{
 	case CSM_CRC32:
-		crc = (uint32_t*)checksum;
-		*crc = crc32(*crc, (unsigned char*)buf, size);
+		*(uint32_t*)checksum = crc32(*(uint32_t*)checksum, (unsigned char*)buf, size);
 		break;
 
 	case CSM_CRC32_0001:
-		crc = (uint32_t*)checksum;
-		*crc = crc32_0001(*crc, (unsigned char*)buf, size);
+		*(uint32_t*)checksum = crc32_0001(*(uint32_t*)checksum, (unsigned char*)buf, size);
+		break;
+
+	case CSM_XXH64:
+		XXH64_update(xxh64_state, buf, size);
 		break;
 
 	case CSM_NONE:
@@ -177,3 +215,45 @@ void update_checksum(unsigned char* checksum, char* buf, int size) {
 	}
 
 }
+
+void finalize_checksum(unsigned char* checksum) {
+
+	switch(cs_mode)
+	{
+	case CSM_XXH64:
+		*(XXH64_hash_t*)checksum = XXH64_digest(xxh64_state);
+		break;
+
+	case CSM_CRC32:
+	case CSM_CRC32_0001:
+	case CSM_NONE:
+		// Nothing to do
+		break;
+	}
+
+}
+
+void release_checksum() {
+    if (xxh64_state != NULL) {
+        XXH64_freeState(xxh64_state);
+        xxh64_state = NULL;
+    }
+}
+
+char* format_checksum(const unsigned char* data, unsigned int size) {
+    if (data == NULL || size == 0) {
+        return NULL;
+    }
+
+    char* hex_string = (char*)malloc(size * 2 + 1);
+    if (hex_string == NULL) {
+        return NULL;
+    }
+
+    for (unsigned int i = 0; i < size; i++) {
+        sprintf(hex_string + (i * 2), "%02x", data[i]);
+    }
+
+    hex_string[size * 2] = '\0';
+    return hex_string;
+}
